@@ -12,6 +12,8 @@ Model Generator for OpenSees ~ solver
 # https://github.com/ioannis-vm/OpenSees_Model_Generator
 
 # pylint: disable=no-member
+# pylint: disable=dangerous-default-value
+# pylint: disable=unused-argument
 
 
 from __future__ import annotations
@@ -32,7 +34,6 @@ import pandas as pd
 import openseespy.opensees as ops  # type: ignore
 import matplotlib.pyplot as plt  # type: ignore
 from .load_case import LoadCase
-from .mesh import subdivide_polygon
 # from .import components
 from .model import Model
 from .ops.element import ElasticBeamColumn
@@ -41,7 +42,7 @@ from .import common
 from .graphics import general_2d
 from .import transformations
 from .collections import Collection
-from .gen.querry import LoadCaseQuerry
+from .gen.query import LoadCaseQuery
 if TYPE_CHECKING:
     from .ops.uniaxial_material import UniaxialMaterial
 
@@ -115,7 +116,7 @@ def test_uniaxial_material(
                 incr_anl = sign * abs(curr_defo - target_defo)
             else:
                 incr_anl = incr * scale[num_subdiv]
-            ops.test('NormDispIncr', norm[num_subdiv],
+            ops.test('EnergyIncr', norm[num_subdiv],
                      steps[num_subdiv], 0)
             ops.algorithm(algor[num_subdiv])
             # ops.integrator("ArcLength", 1.00e1, 1.00e-7)
@@ -214,6 +215,17 @@ class AnalysisStorageSettings:
 @dataclass(repr=False)
 class Analysis:
     """
+    Parent analysis class.
+    Attributes:
+      mdl (Model): a given model
+      load_cases (dict): Dictionary containing load case names and
+        load case objects in which those load cases reside.
+      output_directory (Optional[str]): Where to place the results
+        when it is requested for them to be pickled.
+      silent (bool): Supress printing
+      settings (AnalysisStorageSettings): analysis settings
+      results (Results): analysis results
+      logger(Optional[object]): Logger object
     """
     mdl: Model
     load_cases: dict[str, LoadCase]
@@ -228,7 +240,8 @@ class Analysis:
         """
         Adds a message to the log file
         """
-        if self.settings.log_file:
+        if self.logger:
+            # logger might not have been initialized yet
             self.logger.info(msg)  # type: ignore
 
     def print(self, thing: Any):
@@ -237,9 +250,18 @@ class Analysis:
         """
         if not self.silent:
             print(thing)
+        if self.logger:
+            # logger might not have been initialized yet
+            self.log(thing)
 
     def _init_results(self):
 
+        # initialize output directory
+        if self.output_directory and not os.path.exists(self.output_directory):
+            os.makedirs(
+                self.output_directory,
+                exist_ok=True)
+
         # initialize logger
         if self.settings.log_file:
             logging.basicConfig(
@@ -249,12 +271,6 @@ class Analysis:
             self.logger = logging.getLogger('OpenSees_Model_Generator')
             self.logger.setLevel(logging.DEBUG)
 
-        # initialize output directory
-        if self.output_directory and not os.path.exists(self.output_directory):
-            os.makedirs(
-                self.output_directory,
-                exist_ok=True)
-
         if self.settings.pickle_results and not self.output_directory:
             raise ValueError('Speficy an output directory for the results.')
 
@@ -379,27 +395,37 @@ class Analysis:
         # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
 
         # keep track of defined elements
-        defined_sections = {}
-        defined_materials = {}
+        defined_sections: dict[int, object] = {}
+        defined_materials: dict[int, object] = {}
 
         elms = self.mdl.dict_of_disp_beamcolumn_elements().values()
 
+        def define_material(mat, defined_materials):
+            """
+            A cute recursive function that defines materials with
+            predecessors.
+            """
+            # if the actual material has not been defined yet,
+            if mat.uid not in defined_materials:
+                while hasattr(mat, 'predecessor') \
+                      and mat.predecessor.uid \
+                      not in defined_materials:
+                    # define predecessor
+                    define_material(mat.predecessor, defined_materials)
+                # and also define the actual material
+                ops.uniaxialMaterial(*mat.ops_args())
+                defined_materials[mat.uid] = mat
+
         for elm in elms:
             sec = elm.section
             parts = sec.section_parts.values()
-            n_x = elm.section.n_x
-            n_y = elm.section.n_y
             if sec.uid not in defined_sections:
                 ops.section(*sec.ops_args())
                 defined_sections[sec.uid] = sec
                 for part in parts:
                     mat = part.ops_material
-                    if mat.uid not in defined_materials:
-                        ops.uniaxialMaterial(
-                            *mat.ops_args())
-                        defined_materials[mat.uid] = mat
-                    pieces = subdivide_polygon(
-                        part.outside_shape, part.holes, n_x, n_y)
+                    define_material(mat, defined_materials)
+                    pieces = part.cut_into_tiny_little_pieces()
                     for piece in pieces:
                         area = piece.area
                         z_loc = piece.centroid.x
@@ -408,7 +434,6 @@ class Analysis:
                                   z_loc,
                                   area,
                                   part.ops_material.uid)
-                    plt.show()
             ops.beamIntegration(*elm.integration.ops_args())
             ops.geomTransf(*elm.geomtransf.ops_args())
             ops.element(*elm.ops_args())
@@ -422,14 +447,27 @@ class Analysis:
         # define zerolength elements
         for elm in elms:
             for mat in elm.mats:
-                if mat.uid not in defined_materials:
-                    ops.uniaxialMaterial(
-                        *mat.ops_args())
-                    defined_materials[mat.uid] = mat
+                define_material(mat, defined_materials)
             ops.element(*elm.ops_args())
             defined_elements[elm.uid] = elm
 
-        # node constraints
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
+        # TwoNodeLink element definition #
+        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
+
+        elms = self.mdl.list_of_twonodelink_elements()
+
+        # define twonodelink elements
+        for elm in elms:
+            for mat in elm.mats:
+                define_material(mat, defined_materials)
+            ops.element(*elm.ops_args())
+            defined_elements[elm.uid] = elm
+
+        # ~~~~~~~~~~~~~~~~ #
+        # node constraints #
+        # ~~~~~~~~~~~~~~~~ #
+
         for uid in parent_nodes:
             lvl = self.mdl.levels[parent_node_to_lvl[uid]]
             nodes = lvl.nodes.values()
@@ -440,20 +478,21 @@ class Analysis:
     def _define_loads(self, case_name):
         ops.timeSeries('Linear', 1)
         ops.pattern('Plain', 1, 1)
-        for elm in self.mdl.list_of_beamcolumn_elements():
-            if elm.visibility.skip_opensees_definition:
-                continue
-            udl_total = (self.load_cases[case_name]
-                         .line_element_udl[elm.uid].val)
-            ops.eleLoad('-ele', elm.uid,
-                        '-type', '-beamUniform',
-                        udl_total[1],
-                        udl_total[2],
-                        udl_total[0])
+        # for elm in self.mdl.list_of_beamcolumn_elements():
+        #     if elm.visibility.skip_opensees_definition:
+        #         continue
+        #     udl_total = (self.load_cases[case_name]
+        #                  .line_element_udl[elm.uid].val)
+        #     if not np.isclose(np.sqrt(udl_total @ udl_total), 0.00):
+        #         ops.eleLoad('-ele', elm.uid,
+        #                     '-type', '-beamUniform',
+        #                     udl_total[1],
+        #                     udl_total[2],
+        #                     udl_total[0])
 
-        for node in self.mdl.list_of_all_nodes():
-            ops.load(node.uid, *self.load_cases[case_name]
-                     .node_loads[node.uid].val)
+        # for node in self.mdl.list_of_all_nodes():
+        #     ops.load(node.uid, *self.load_cases[case_name]
+        #              .node_loads[node.uid].val)
 
     ####################################################
     # Methods that read back information from OpenSees #
@@ -560,7 +599,9 @@ class Analysis:
             nodes,
             elastic_beamcolumn_elements,
             disp_beamcolumn_elements,
-            zerolength_elements):
+            zerolength_elements,
+            custom_read_results_method,
+            custom_read_results_method_args):
         self._read_node_displacements(case_name, step, nodes)
         self._read_node_velocities(case_name, step, nodes)
         self._read_node_accelerations(case_name, step, nodes)
@@ -570,15 +611,15 @@ class Analysis:
             self._read_frame_element_forces(
                 case_name,
                 step, elastic_beamcolumn_elements)
-            self._read_frame_element_forces(
-                case_name,
-                step, disp_beamcolumn_elements)
         # if self.settings.store_fiber:
         #     self._read_frame_fiber_stress_strain()
         if self.settings.store_release_force_defo:
             self._read_release_moment_rot(
                 case_name,
                 step, zerolength_elements)
+        if custom_read_results_method is not None:
+            custom_read_results_method(
+                **custom_read_results_method_args)
 
     ##################################
     # Numeric Result Post-processing #
@@ -619,7 +660,9 @@ class StaticAnalysis(Analysis):
     Static analysis.  Stores all results (for each load case) in one
     single step.
     """
-    def run(self):
+    def run(self,
+            custom_read_results_method=None,
+            custom_read_results_method_args={}):
         """
         Runs the static analysis.
         """
@@ -640,7 +683,7 @@ class StaticAnalysis(Analysis):
             ops.system(LN_ANALYSIS_SYSTEM)
             ops.numberer(NUMBERER)
             ops.constraints(*CONSTRAINTS)
-            ops.test('NormDispIncr', 1.0e-8, 20, 3)
+            ops.test('EnergyIncr', 1.0e-8, 20, 3)
             ops.algorithm('Newton')
             ops.integrator('LoadControl', 1.0)
             ops.analysis('Static')
@@ -651,7 +694,9 @@ class StaticAnalysis(Analysis):
                 nodes,
                 elastic_elems,
                 disp_elems,
-                zerolength_elems)
+                zerolength_elems,
+                custom_read_results_method,
+                custom_read_results_method_args)
             if self.settings.pickle_results:
                 self._write_results_to_disk()
 
@@ -683,97 +728,100 @@ class ModalAnalysis(Analysis):
         # to overcome this, we run a separate analysis imposing those
         # displacements to get the element forces.
 
-        # note: work in progress. currently this doesn't fully work
+        for step in range(self.num_modes):
+            for elm in elems:
+                assert isinstance(elm, ElasticBeamColumn)
+                # displacements at the two ends (global system)
+                u_i = (self.results[case_name].node_displacements
+                       [elm.nodes[0].uid][step][0:3])
+                r_i = (self.results[case_name].node_displacements
+                       [elm.nodes[0].uid][step][3:6])
+                u_j = (self.results[case_name].node_displacements
+                       [elm.nodes[1].uid][step][0:3])
+                r_j = (self.results[case_name].node_displacements
+                       [elm.nodes[1].uid][step][3:6])
+                offset_i = elm.geomtransf.offset_i
+                offset_j = elm.geomtransf.offset_j
+                u_i_o = transformations.offset_transformation(
+                    offset_i, np.array(u_i), np.array(r_i))
+                u_j_o = transformations.offset_transformation(
+                    offset_j, np.array(u_j), np.array(r_j))
 
-        for case_name in self.load_cases:
-            for step in range(self.num_modes):
-                for elm in elems:
-                    # if step == 3:
-                    #     import pdb
-                    #     pdb.set_trace()
-                    assert isinstance(elm, ElasticBeamColumn)
-                    # displacements at the two ends (global system)
-                    u_i = (self.results[case_name].node_displacements
-                           [elm.nodes[0].uid][step][0:3])
-                    r_i = (self.results[case_name].node_displacements
-                           [elm.nodes[0].uid][step][3:6])
-                    u_j = (self.results[case_name].node_displacements
-                           [elm.nodes[1].uid][step][0:3])
-                    r_j = (self.results[case_name].node_displacements
-                           [elm.nodes[1].uid][step][3:6])
-                    offset_i = elm.geomtransf.offset_i
-                    offset_j = elm.geomtransf.offset_j
-                    u_i_o = transformations.offset_transformation(
-                        offset_i, np.array(u_i), np.array(r_i))
-                    u_j_o = transformations.offset_transformation(
-                        offset_j, np.array(u_j), np.array(r_j))
+                x_vec = elm.geomtransf.x_axis
+                y_vec = elm.geomtransf.y_axis
+                z_vec: nparr = np.cross(x_vec, y_vec)
 
-                    x_vec = elm.geomtransf.x_axis
-                    y_vec = elm.geomtransf.y_axis
-                    z_vec: nparr = np.cross(x_vec, y_vec)
+                # global -> local transformation matrix
+                transf_global2local = \
+                    transformations.transformation_matrix(
+                        x_vec, y_vec, z_vec)
+                u_i_local = transf_global2local @ u_i_o
+                r_i_local = transf_global2local @ r_i
+                u_j_local = transf_global2local @ u_j_o
+                r_j_local = transf_global2local @ r_j
 
-                    # global -> local transformation matrix
-                    transf_global2local = \
-                        transformations.transformation_matrix(
-                            x_vec, y_vec, z_vec)
-                    u_i_local = transf_global2local @ u_i_o
-                    r_i_local = transf_global2local @ r_i
-                    u_j_local = transf_global2local @ u_j_o
-                    r_j_local = transf_global2local @ r_j
+                # # element UDL
+                udl = (self.load_cases[case_name]
+                       .line_element_udl[elm.uid].val)
+                # note: modal analsis doesn't account for applied loads.
+                # this will cause issues with plotting if loads
+                # have been applied.
+                if np.linalg.norm(udl) > common.EPSILON:
+                    raise ValueError('Loads applied at modal load case.')
 
-                    # # element UDL
-                    udl = (self.load_cases[case_name]
-                           .line_element_udl[elm.uid].val)
-                    # note: modal analsis doesn't account for applied loads.
-                    # this will cause issues with plotting if loads
-                    # have been applied.
-                    if np.linalg.norm(udl) > common.EPSILON:
-                        raise ValueError('Loads applied at modal load case.')
+                # stiffness matrix terms
+                length = elm.clear_length()
+                etimesa = elm.section.e_mod * elm.section.area
+                etimesi_maj = elm.section.e_mod * elm.section.i_x
+                etimesi_min = elm.section.e_mod * elm.section.i_y
+                gtimesj = elm.section.g_mod * elm.section.j_mod
 
-                    # stiffness matrix terms
-                    length = elm.clear_length()
-                    etimesa = elm.section.e_mod * elm.section.area
-                    etimesi_maj = elm.section.e_mod * elm.section.i_x
-                    # eimin = elm.section.e_mod * elm.section.Iy
-                    gtimesj = elm.section.g_mod * elm.section.j_mod
+                deformation_vector = np.concatenate(
+                    (u_i_local, r_i_local, u_j_local, r_j_local))
 
-                    # deformations
-                    d_l = u_j_local[0] - u_i_local[0]
-                    theta_tor = r_j_local[0] - r_i_local[0]
-                    du_xy = u_j_local[1] - u_i_local[1]
-                    # du_xz = u_j_local[2] - u_i_local[2]
+                # axial load
+                n_i = np.array([
+                    [etimesa/length, -etimesa/length]
+                ]) @ deformation_vector[[0, 6]]
 
-                    # axial load
-                    n_i = etimesa / length * d_l
+                # torsion
+                t_i = np.array([
+                    [gtimesj/length, -gtimesj/length]
+                ]) @ deformation_vector[[3, 9]]
 
-                    # strong axis bending
-                    mz_i = \
-                        (4.00 * etimesi_maj / length * r_i_local[2]
-                         + 2.00 * etimesi_maj / length * r_j_local[2]
-                         - 6.00 * etimesi_maj / length**2 * du_xy)
-                    qy_i = (6.00 * etimesi_maj / length**2
-                            * (r_i_local[2] + r_j_local[2])
-                            - 12.00 * etimesi_maj / length**3 * du_xy)
+                # major shear and minor bending
+                f3_m2 = np.array([
+                    [12.00*etimesi_min/length**3,
+                     -6.00*etimesi_min/length**2,
+                     -12.00*etimesi_min/length**3,
+                     -6.00*etimesi_min/length**2],
+                    [-6.00*etimesi_min/length**2,
+                     4.00*etimesi_min/length,
+                     6.00*etimesi_min/length**2,
+                     2.00*etimesi_min/length]
+                ]) @ deformation_vector[[2, 4, 8, 10]]
 
-                    # # weak axis bending
-                    # # work in progress...
-                    # myi = \
-                    #     -(4.00 * eimin / length * (-r_i_local[1])
-                    #      + 2.00 * eimin / length * (-r_j_local[1])
-                    #      - 6.00 * eimin / length * (du_xz))
-                    # qzi = (6.00 * eimin / length**2
-                    #        * ((-r_i_local[2]) + (-r_j_local[2]))
-                    #        - 12.00 * eimin / length**3 * (du_xz))
-                    myi = 0.00
-                    qzi = 0.00
+                # minor shear and major bending
+                f2_m3 = np.array([
+                    [12.00*etimesi_maj/length**3,
+                     6.00*etimesi_maj/length**2,
+                     -12.00*etimesi_maj/length**3,
+                     6.00*etimesi_maj/length**2],
+                    [6.00*etimesi_maj/length**2,
+                     4.00*etimesi_maj/length,
+                     -6.00*etimesi_maj/length**2,
+                     2.00*etimesi_maj/length]
+                ]) @ deformation_vector[[1, 5, 7, 11]]
 
-                    # torsion
-                    t_i = gtimesj / length * theta_tor
+                forces_vector_local = np.array((
+                    n_i[0], f2_m3[0], f3_m2[0],
+                    t_i[0], f3_m2[1], f2_m3[1]))
 
-                    # store results
-                    (self.results[case_name].element_forces
-                     [elm.uid][step]) = \
-                        np.array((n_i, qy_i, qzi, t_i, myi, mz_i))
+                # store results
+                (self.results[case_name].element_forces
+                 # [elm.uid][step]) = \
+                 #    np.array((n_i, qy_i, qzi, t_i, myi, mz_i))
+                 [elm.uid][step]) = forces_vector_local
 
     def run(self):
         """
@@ -841,7 +889,7 @@ class NonlinearAnalysis(Analysis):
         ops.system(system)
         ops.numberer(NUMBERER)
         ops.constraints(*CONSTRAINTS)
-        ops.test('NormDispIncr', 1.0e-6, 100, 3)
+        ops.test('EnergyIncr', 1.0e-6, 100, 3)
         ops.algorithm('RaphsonNewton')
         ops.integrator('LoadControl', 1)
         ops.analysis('Static')
@@ -849,62 +897,88 @@ class NonlinearAnalysis(Analysis):
         if check != 0:
             raise ValueError('Analysis Failed')
 
-    # def _acceptance_criteria(self):
-    #     for elm in self.mdl.list_of_line_elements():
+    # failed attempt
+    # def remove_damaged_elements(self):
+    #     """
+    #     Applies to fiber-section elements that use the Fatigue
+    #     material to simulate damage.
+    #     Elements for which any section has Damage >= 1 to more than a
+    #     set percentage of its fibers are removed.
+    #     Assumes that the section consists of a single part called 'main'
+    #     """
+    #     max_dmg_ratio = 0.30
+    #     candidate_elms = []
 
-    #         if elm.model_as['type'] != 'elastic':
-    #             continue
-    #         mat = elm.section.material
-    #         if mat.name == 'steel':
-    #             capacity_t = mat.parameters['Fy']/mat.parameters['E0']
-    #             capacity_c = -capacity_t
-    #         else:
-    #             raise ValueError('Unsupported material')
-    #         strains = []
-    #         x_vec = elm.x_axis
-    #         y_vec = elm.y_axis
-    #         z_vec = elm.z_axis
-    #         trans_global2local: nparr = np.vstack((x_vec, y_vec, z_vec))
-    #         forces_global: nparr = np.array(
-    #             self.element_forces[str(elm.uid)][-1][0:3])
-    #         moments_global_ends: nparr = np.array(
-    #             self.element_forces[str(elm.uid)][-1][3:6])
+    #     for elm in self.mdl.dict_of_disp_beamcolumn_elements().values():
+    #         part = elm.section.section_parts['main']
+    #         if isinstance(elm.section, FiberSection):
+    #             if isinstance(part.ops_material, Fatigue):
+    #                 candidate_elms.append(elm)
 
-    #         moments_global_clear = \
-    #             transformations.offset_transformation(
-    #                 elm.offset_i, moments_global_ends, forces_global)
+    #     for elm in candidate_elms:
+    #         part = elm.section.section_parts['main']
+    #         mat_uid = part.ops_material.uid
+    #         sec_dmg_percentages = []
+    #         for sec_num in range(1, elm.integration.n_p+1):
+    #             num_damaged_fibers = 0
+    #             pieces = subdivide_polygon(
+    #                 part.outside_shape, part.holes,
+    #                 elm.section.n_x, elm.section.n_y)
+    #             for piece in pieces:
+    #                 z_loc = piece.centroid.x
+    #                 y_loc = piece.centroid.y
+    #                 dmg = ops.eleResponse(
+    #                     elm.uid, 'section', str(sec_num),
+    #                     'fiber', str(y_loc), str(z_loc), str(mat_uid),
+    #                     'damage')[0]
+    #                 if dmg > 1.00:
+    #                     num_damaged_fibers += 1
+    #             damage_percentage = float(
+    #                 num_damaged_fibers)/float(len(pieces))
+    #             sec_dmg_percentages.append(damage_percentage)
+    #         max_damage = max(sec_dmg_percentages)
+    #         if max_damage > max_dmg_ratio:
+    #             # need to remove that element
+    #             import pdb
+    #             pdb.set_trace()
+    #             print('ok')
+    #             ops.remove('ele', elm.uid)
+    #             ops.reset()
 
-    #         ni, qyi, qzi = trans_global2local @ forces_global
-    #         ti, myi, mzi = trans_global2local @ moments_global_clear
-
-    #         wx, wy, wz = elm.udl.total()
-
-    #         len_clr = elm.length_clear
-    #         t = np.linspace(0.00, len_clr, num=9)
-
-    #         nx_vec = - t * wx - ni
-    #         mz_vec = t**2 * 0.50 * wy + t * qyi - mzi
-    #         my_vec = t**2 * 0.50 * wz + t * qzi + myi
-
-    #         prop = elm.section.mesh.geometric_properties()
-    #         area = prop['area']
-    #         iy = prop['inertia']['ixx']
-    #         iz = prop['inertia']['iyy']
-    #         young_mod = elm.section.material.parameters['E0']
-
-    #         for val in elm.section.snap_points.values():
-    #             z, y = val
-    #             stress = nx_vec/area \
-    #                 + my_vec/iz * z \
-    #                 - mz_vec/iy * y
-    #             strain = stress / young_mod
-    #             strains.extend(strain)
-    #         emax = np.max(np.array(strains))
-    #         emin = np.min(np.array(strains))
-    #         if ((emax > capacity_t) or (emin < capacity_c)):
-    #             raise ValueError(
-    #                 "Acceptance criteria failed for element " +
-    #                 str(elm.uid))
+    # def intervention(self, n_steps_success):
+    #     if n_steps_success == -1:
+    #     # if n_steps_success == 2000:
+    #         elm = self.mdl.dict_of_disp_beamcolumn_elements()[49]
+    #         part = elm.section.section_parts['main']
+    #         mat_uid = part.ops_material.uid
+    #     else:
+    #         return
+    #     for sec_num in range(1, elm.integration.n_p+1):
+    #         zs = []
+    #         ys = []
+    #         dmgs = []
+    #         pieces = subdivide_polygon(
+    #             part.outside_shape, part.holes,
+    #             elm.section.n_x, elm.section.n_y)
+    #         for piece in pieces:
+    #             z_loc = piece.centroid.x
+    #             y_loc = piece.centroid.y
+    #             dmg = ops.eleResponse(
+    #                 elm.uid, 'section', str(sec_num),
+    #                 'fiber', str(y_loc), str(z_loc),
+    #                 str(mat_uid), 'damage')[0]
+    #             zs.append(z_loc)
+    #             ys.append(y_loc)
+    #             dmgs.append(dmg)
+    #         print(dmgs)
+    #         plt.scatter(zs, ys, s=0.1)
+    #         for i, txt in enumerate(dmgs):
+    #             plt.annotate(f'{txt:.1f}', (zs[i], ys[i]))
+    #         plt.title(f'Section {sec_num}/{sec_num}')
+    #         plt.show()
+    #     import pdb
+    #     pdb.set_trace()
+    #     print()
 
     def retrieve_node_displacement(self, uid, case_name):
         """
@@ -1019,8 +1093,8 @@ class PushoverAnalysis(NonlinearAnalysis):
     """
     def _apply_lateral_load(
             self, case_name, direction, modeshape=None, node=None):
-        querry = LoadCaseQuerry(self.mdl, self.load_cases[case_name])
-        distribution = querry.level_masses()
+        query = LoadCaseQuery(self.mdl, self.load_cases[case_name])
+        distribution = query.level_masses()
         distribution = distribution / np.linalg.norm(distribution)
 
         # define the load pattern
@@ -1067,7 +1141,9 @@ class PushoverAnalysis(NonlinearAnalysis):
                                modeshape_ampl[i]))
 
     def run(self, direction, target_displacements,
-            control_node, displ_incr, modeshape=None, loaded_node=None):
+            control_node, displ_incr, modeshape=None, loaded_node=None,
+            custom_read_results_method=None,
+            custom_read_results_method_args={}):
         """
         Run pushover analysis
         Arguments:
@@ -1085,6 +1161,7 @@ class PushoverAnalysis(NonlinearAnalysis):
             incremental load is applied entirely on that node.
             Otherwise, the incremental loads are distributed to all nodes.
         """
+        self.log(f'Direction: {direction}')
         if direction == 'x':
             control_dof = 0
         elif direction == 'y':
@@ -1094,9 +1171,11 @@ class PushoverAnalysis(NonlinearAnalysis):
         else:
             raise ValueError("Direction can be 'x', 'y' or 'z'")
 
+        self.log('Initializing containers')
         self._init_results()
 
         for case_name in self.load_cases:
+            self.log(f'Load case: {case_name}')
             nodes = self.mdl.list_of_all_nodes()
             nodes.extend(self.load_cases[case_name].parent_nodes.values())
             elastic_elems = [
@@ -1107,8 +1186,11 @@ class PushoverAnalysis(NonlinearAnalysis):
                 if not elm.visibility.skip_opensees_definition]
             zerolength_elems = self.mdl.list_of_zerolength_elements()
 
+            self.log('Defining elements in OpenSees')
             self._to_opensees_domain(case_name)
+            self.log('Defining loads')
             self._define_loads(case_name)
+            self.log('Running gravity analysis')
             self._run_gravity_analysis(NL_ANALYSIS_SYSTEM)
 
             curr_displ = ops.nodeDisp(control_node.uid, control_dof+1)
@@ -1119,8 +1201,11 @@ class PushoverAnalysis(NonlinearAnalysis):
                 nodes,
                 elastic_elems,
                 disp_elems,
-                zerolength_elems)
+                zerolength_elems,
+                custom_read_results_method,
+                custom_read_results_method_args)
 
+            self.log('Starting pushover analysis')
             ops.wipeAnalysis()
             ops.loadConst('-time', 0.0)
             self._apply_lateral_load(
@@ -1132,9 +1217,9 @@ class PushoverAnalysis(NonlinearAnalysis):
             num_subdiv = 0
             num_times = 0
 
-            scale = [1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]
-            steps = [25, 50, 75, 100, 250, 500]
-            norm = [1.0e-6] * 5 + [1.0e-2]
+            scale = [1.0, 0.1, 0.01]
+            steps = [25, 50, 100]
+            norm = [1.0e-6, 1.0e-1, 1.0e-1]
 
             try:
 
@@ -1171,12 +1256,16 @@ class PushoverAnalysis(NonlinearAnalysis):
                         ops.analysis("Static")
                         flag = ops.analyze(1)
                         if flag != 0:
-                            # analysis failed
                             if num_subdiv == len(scale) - 1:
                                 # can't refine further
                                 self.print('===========================')
                                 self.print('Analysis failed to converge')
                                 self.print('===========================')
+                                if self.logger:
+                                    self.logger.warning(  # type: ignore
+                                        "Analysis failed"  # type: ignore
+                                        f" at disp {curr_displ:.5f}"
+                                    )  # type: ignore
                                 total_fail = True
                                 break
                             # can still reduce step size
@@ -1194,8 +1283,9 @@ class PushoverAnalysis(NonlinearAnalysis):
                                 nodes,
                                 elastic_elems,
                                 disp_elems,
-                                zerolength_elems)
-                            # self._acceptance_criteria()
+                                zerolength_elems,
+                                custom_read_results_method,
+                                custom_read_results_method_args)
                             curr_displ = ops.nodeDisp(
                                 int(control_node.uid), control_dof+1)
                             print('Target displacement: '
@@ -1216,7 +1306,9 @@ class PushoverAnalysis(NonlinearAnalysis):
                 nodes,
                 elastic_elems,
                 disp_elems,
-                zerolength_elems)
+                zerolength_elems,
+                custom_read_results_method,
+                custom_read_results_method_args)
             self.print(f'Number of saved analysis steps: {n_steps_success}')
             metadata = {'successful steps': n_steps_success}
             self.results[case_name].n_steps_success = n_steps_success
@@ -1374,7 +1466,9 @@ class NLTHAnalysis(NonlinearAnalysis):
             finish_time=0.00,
             skip_steps=1,
             damping={'type': None},
-            print_progress=True):
+            print_progress=True,
+            custom_read_results_method=None,
+            custom_read_results_method_args={}):
         """
         Run the nonlinear time-history analysis
         Args:
@@ -1469,7 +1563,9 @@ class NLTHAnalysis(NonlinearAnalysis):
             nodes,
             elastic_elems,
             disp_elems,
-            zerolength_elems
+            zerolength_elems,
+            custom_read_results_method,
+            custom_read_results_method_args
         )
 
         # time-history analysis
@@ -1519,7 +1615,7 @@ class NLTHAnalysis(NonlinearAnalysis):
                 'modal damping defined')
 
         self.log('Starting transient analysis')
-        ops.test('NormDispIncr', 1e-6, 50, 0)
+        ops.test('EnergyIncr', 1e-6, 50, 0)
         ops.algorithm("KrylovNewton")
         # ops.integrator('Newmark', 0.50, 0.25)
         ops.integrator('TRBDF2')
@@ -1628,18 +1724,17 @@ class NLTHAnalysis(NonlinearAnalysis):
                 speed = total_step_count / (time.perf_counter() - now)
                 if total_step_count % 50 == 0:
                     # provide run speed statistics
-                    # debug
                     print(f'Average speed: {speed:.2f} steps/s')
-                    self.log(f'Average speed: {speed:.2f} steps/s')
                 if check != 0:
                     # analysis failed
                     if num_subdiv == len(scale) - 1:
                         self.print('===========================')
                         self.print('Analysis failed to converge')
                         self.print('===========================')
-                        self.logger.warning(  # type: ignore
-                            "Analysis failed"  # type: ignore
-                            f" at time {curr_time:.5f}")  # type: ignore
+                        if self.logger:
+                            self.logger.warning(  # type: ignore
+                                "Analysis failed"  # type: ignore
+                                f" at time {curr_time:.5f}")  # type: ignore
                         analysis_failed = True
                         break
                     # can still reduce step size
@@ -1659,7 +1754,9 @@ class NLTHAnalysis(NonlinearAnalysis):
                             nodes,
                             elastic_elems,
                             disp_elems,
-                            zerolength_elems
+                            zerolength_elems,
+                            custom_read_results_method,
+                            custom_read_results_method_args
                         )
                         self.time_vector.append(curr_time)
                     if print_progress:
@@ -1674,7 +1771,8 @@ class NLTHAnalysis(NonlinearAnalysis):
 
         except KeyboardInterrupt:
             self.print("Analysis interrupted")
-            self.logger.warning("Analysis interrupted")  # type: ignore
+            if self.logger:
+                self.logger.warning("Analysis interrupted")  # type: ignore
 
         metadata = {'successful steps': n_steps_success,
                     'analysis_finished_successfully': not analysis_failed}
